; 7-ade decompression wrapper
; ©2016 Daniel Beardsmore
; http://telcontar.net/Software/7-ade

#include <Misc>

runExtract(toExtract) {
	; Extract the specified archive
	; One or more temporary directories may be left behind if an error occurs:
	; this is because FileRemoveDir is *never* called with the recursion option, in case an unforeseen
	; error leads to a request to delete a valid directory
	
	SplitPath, toExtract, archiveName, archiveDir, , archiveNameStem
	
	ruleset := findRulesetMatching(toExtract)
	dbg("Ruleset " . ruleset["name"] . " selected for " . toExtract)
	
	deleteWhenDone := false
	tempOutDir := ""
	args := ""
	
	if (ruleset["target path"])
		archiveDir := ruleset["target path"]
	
	if (ruleset["overwrite"])
		args .= "-aoa"
	
	; Handle tarballs, where we first need to extract the archive contents into a temporary directory,
	; and perform a second expansion operation on the inner item
	if (isWrapped(archiveName)) {
		tempOutDir := makeTempDir()
		innerName := getInnerName(toExtract)
		newPath := tempOutDir . "\" . innerName
		
		r := call7zG(toExtract, "", tempOutDir)
		if (r == "ERROR") {
			msgBox 16, %k_APPNAME%, Attempting to invoke 7-Zip to perform an extraction operation failed.
			return false
		}
		else if (r != 0 && !FileExist(newPath)) {
			msgBox 48, %k_APPNAME%, % "Extracting " . archiveName . " should have yielded " . innerName . " in " . tempOutDir . "; this file did not materialise."
			return false
		}
		
		toExtract := newPath
		deleteWhenDone := true
	}
	
	rootItemName := ""
	needEnclosingFolder := isEnclosingFolderNeeded(toExtract, rootItemName)
	
	if (needEnclosingFolder) {
		; Archive has multiple items at its root -- create a new directory to contain these items, using the name of the archive itself
		
		dbg("Expansion mode: enclosing folder")
		
		archiveDir .= "\" . archiveNameStem
		if (FileExist(archiveDir) && !ruleset["overwrite"]) {
			archiveDir := getUniqueName(archiveDir, "D")
		}
		FileCreateDir, %archiveDir%
		
		r := call7zG(toExtract, args, archiveDir)
	}
	else if (FileExist(archiveDir . "\" . rootItemName) && !ruleset["overwrite"]) {
		; Archive has a single item at the root, an item of this name already exists, and overwrite mode is not enabled
		
		dbg("Expansion mode: duplicate item")
		
		tempDir := makeTempDir(archiveDir)
		r := call7zG(toExtract, args, tempDir)
		if (r == 0) {
			extractedItem := tempDir . "\" . rootItemName
			itemCheck := FileExist(extractedItem)
			if (itemCheck) {
				if (instr(itemCheck, "D")) {
					newName := getUniqueName(archiveDir . "\" . rootItemName, "D")
					FileMoveDir, %extractedItem%, %newName%, R
				}
				else {
					newName := getUniqueName(archiveDir . "\" . rootItemName, "F")
					FileMove, %extractedItem%, %newName%
				}
				FileRemoveDir, %tempDir%
			}
			else {
				msgBox 48, %k_APPNAME%, % "Extracting " . extractedItem . " has not worked; this file does not exist when it should."
			}
		}
		else if (r == "ERROR") {
			FileRemoveDir, %tempDir%
		}
	}
	else {
		dbg("Expansion mode: basic")
		
		r := call7zG(toExtract, args, archiveDir)
	}
	
	if (r == "ERROR") {
		msgBox 16, %k_APPNAME%, "Attempting to run 7zG.exe failed."
	}
	
	if (deleteWhenDone && tempOutDir) {
		if (FileExist(toExtract))
			FileDelete, %toExtract%
		FileRemoveDir, %tempOutDir%, 0
	}
}


call7zG(archivePath, args, dir) {
	; Extract an archive using 7zG.exe, with specified command-line arguments (can be left blank) into the specified directory
	
	if (args)
		args := " " . args
	
	cmd := """" . g_7ZipPath . "\7zG.exe"" x """ . archivePath . """" . args . " -o""" . dir . """"
	
	dbg("To extract: " . archivePath)
	dbg("Write to: " . dir)
	dbg("Command: " . cmd)
	
	RunWait, % cmd, , UseErrorLevel
	
	return ErrorLevel
}


getInnerName(archive) {
	; Get the name of the tar file within a gzip archive
	; Don't allow slashes (forward or backward) in the name to avoid ending up in another directory
	; Also, if the gzip contains more than one file, either the universe is broken, or we're not parsing the contents correctly
	
	items := getArchiveContents(archive)
	
	if (items.MaxIndex() != 1) {
		msgBox 16, %k_APPNAME%: Bad archive, % "Tarball " . archive . " seemingly contains " . items.MaxIndex() . " item(s), when it should contain only 1."
		return false
	}
	
	if (InStr(items[1], "\")) {
		msgBox 16, %k_APPNAME%: Bad archive, % "Tarball " . archive . " contains internal name " . items[1] . " with a '\'; this is not supported."
		return false
	}
	
	if (InStr(items[1], "/")) {
		msgBox 16, %k_APPNAME%: Bad archive, % "Tarball " . archive . " contains internal name " . items[1] . " with a '/'; this is not supported."
		return false
	}
	
	return items[1]
}


isEnclosingFolderNeeded(archive, byref rootItem) {
	items := getArchiveContents(archive)
	
	rootItems := 0
	needEnclosingFolder := false
	lastRootItem := ""
	
	for index, itemName in items {
		IfInString itemName, \
			itemName := SubStr(itemName, 1, InStr(itemName, "\") - 1)
		
		if (itemName != lastRootItem) {
			rootItems++
			lastRootItem := itemName
		}
		
		if (rootItems > 1) {
			needEnclosingFolder := true
			break
		}
	}
	
	if (rootItems == 1)
		rootItem := lastRootItem
	
	return needEnclosingFolder
}


getArchiveContents(archivePath) {
	; Returns a list of all item names within the archive
	; Maybe at some point we can include some exit conditions to save building a list of RAM of more items than we need
	
	cmd := """" . g_7ZipPath . "\7z.exe"" l """ . archivePath . """ -sccUTF-8"
	exitCode := -1
	
	outText := StdoutToVar_CreateProcess(cmd, , , exitCode)
	outText := Trim(outText, "`r`n")
	StringReplace, outText, outText, `r`n`r`n, `r`n, All
	
	if (exitCode != 0)
		die(k_EXIT_INVOKEERROR, "Output from 7-Zip:`r`n`r`n" . outText . "`r`n`r`nCommand:`r`n" . cmd)
	
	lineEnd := Chr(13) . Chr(10)
	inList := false
	items := []
	
	Loop, Parse, outText, %lineEnd%
	{
		if (A_LoopField == "")
			Continue
		
		if (SubStr(A_LoopField, 1, 5) == "-----")
		{
			if (inList)
				break
			else
			{
				inList := true
				Continue
			}
		}
		
		if (inList)
		{
			items.insert(SubStr(A_LoopField, 54))
		}
	}
	
	return items
}


findRulesetMatching(toExtract) {
	SplitPath, toExtract, archiveName, archiveDir
	
	archiveNameUc := uc(archiveName)
	archiveDirUc := uc(archiveDir)
	
	for ruleName, rule in g_config["rulesets"]
	{
		hasMatched := false
		
		; Name checks
		if (rule["name is"])
		{
			hasMatched := archiveNameUc == uc(rule["name is"])
			if (!hasMatched)
				continue
		}
		if (rule["name begins"])
		{
			hasMatched := substr(archiveNameUc, 1, strlen(rule["name begins"])) == uc(rule["name begins"])
			if (!hasMatched)
				continue
		}
		
		if (rule["name ends"])
		{
			hasMatched := substr(archiveNameUc, 1 - strlen(rule["name ends"])) == uc(rule["name ends"])
			if (!hasMatched)
				continue
		}
		
		if (rule["name contains"])
		{
			hasMatched := instr(archiveName, rule["name contains"], false) >= 1
			if (!hasMatched)
				continue
		}
		
		; Path checks
		if (rule["path is"])
		{
			hasMatched := archiveDirUc == uc(rule["path is"])
			if (!hasMatched)
				continue
		}
		
		if (rule["path begins"])
		{
			hasMatched := substr(archiveDirUc, 1, strlen(rule["path begins"])) == uc(rule["path begins"])
			if (!hasMatched)
				continue
		}
		
		if (rule["path contains"])
		{
			hasMatched := instr(archiveDir, rule["path contains"], false) >= 1
			if (!hasMatched)
				continue
		}
			
		if (hasMatched)
		{
			rule["name"] := ruleName
			return rule
		}
	}
	
	if (g_config["defaults"])
		return g_config["defaults"]
	else
		return {}
}


isWrapped(archiveName) {
	; If the archive name ends with any of the defined wrapper extensions, then return true
	; In such cases, we need to run an intermediate expansion to remove the wrapper compression
	; Based on the list of extensions recognised by 7-Zip 9.20
	
	wrappers := [ ".tar.gz", ".tgz", ".tar.bzip2", ".tar.bz2", ".tar.bz", ".tar.bz2", ".txz", ".tar.xz" ]
	
	StringLower archiveName, archiveName
	
	for index, extension in wrappers {
		if (substr(archiveName, 1 - strlen(extension)) == extension)
			return true
	}
	return false
}


makeTempDir(placeInto := "") {
	if (!placeInto)
		placeInto := A_Temp
	
	path := getUniqueName(placeInto . "\" . k_APPNAME . "_temp", "D")
	
	FileCreateDir, %path%
	if (ErrorLevel == 1)
		die(k_EXIT_INTERNALERROR, "Creation of temporary directory """ . path . """ failed with code " . A_LastError)
	
	return path
}


getUniqueName(path, type, startCount := 2) {
	; Check whether an item exists with the path specified
	; If so, return a new path with the next available numeric suffix
	; Type should be "D" for directory paths, to append the suffix after the whole name
	; Otherwise, the suffix is added between the name stem and extension
	
	if (!FileExist(path))
		return path
	
	SplitPath, path, itemName, itemDir, itemExt, itemNameStem
	
	timeLimit := 10
	startTime := A_Now
	duplicateNo := startCount
	
	loop {
		if (A_Now - startTime > timeLimit)
			die(k_EXIT_INTERNALERROR
				, "Attempting to form a unique pathname for " . path . " has reached the deadline of " . timeLimit . " seconds after " . (duplicateNo - startCount) . " iterations. Something seems to be wrong")
				
		if (type == "D" || !itemExt)
			tryName := itemName . " (" . duplicateNo . ")"
		else
			tryName := itemNameStem . " (" . duplicateNo . ")." . itemExt
		
		tryPath := itemDir . "\" . tryName
		duplicateNo++
		
		if (!FileExist(tryPath))
			return tryPath
	}
}